/*****************************************************************************
*                                                                            *
*                                 MIDI.C                                     *
*                                                                            *
******************************************************************************
Author:
    Joe Little    (with portions shamelessly copied from Tom Coonans' programs)

Version:
    D1.2  04-05-98 Beefed up MIDI message parse routine.
    D1.1  04-05-98 Cleaned up MIDI message parse routine.
    D1.0  10-01-97 Initial release

Changes:

Description:

    PIC based MIDI light dimmer controller
    The naming convention used here is some confusing.
    A MIDI Channel refers to 1 of 16 possible sets of 127 notes or controls 
    A MIDI Control refers to 1 of 127 possible notes of a particular channel.
    A DIMMER PACK contains channels; 4 triacs and one relay, labeled A thru E.  
    Dimmer channel A appears at the base address, Dimmer channel E appears at 
    base address + 4.  Channels A - D are Triacs. Channel E is the relay.

    The hardware project associated with this design is set up for 4 channels 
    of lights and one relay.  Each light channel can handle 6 amps, limited by 
    the traces on the PWB. The total the Box can handle is 15 amps.  This means that
    if all channels have 500 watts of lamps, and you turn all channels on full
    bright at the same time, then you blow the fuse.  Do not exceed the power rating
    of the weakest link in the system.  Weak links include the power outlet that you 
    plug into, the extension cord used, the linecord on the dimmer pack, the traces 
    on the board, the heatsink for the triacs, and the filter inductors. For nice 
    round numbers, including a little safety margin, 120V operation limits each channel
    to 500W of AC control. The relay provides a dry contact normally open switch output.
    The relay is rated at 125V one half Amp.

    Midi addressing is two bytes.  Its' divided into 127 controls
    for each of 16 channels. Midi has a three (8 bit) word message per 
    transaction.  First word contains a command and MIDI channel number.  
    Second word has the MIDI control number, and the third word is the 
    controls value.
           
    A MICROCHIP 16C63 Microcontroller is used.  The main clock oscillator
    is a 20.0 MHz crystal.  It doesn't divide easily to even nanosecond chunks 
    for easy time calculations, but 120Hz is not a round number to time with 
    anyway.  

    This program is written in C using the CCS compiler.

    In general, the program is nearly all interrupt driven.  The main loop
    reads the config switches upon powerup to gather addressing and options, 
    but from there on, there's nothing much to do except respond to AC zero
    crossings, serial MIDI charactors, and timer ticks.

    A timer and counter is created to provide interrupts every 1/127th of a 60Hz
    AC half cycle.   0 to 100% intensity on the lamp corresponds to a 0 to 127
    command from MIDI.  AC power is passed to the lamp thru a TRIAC.  If you
    wait until half the AC half cycle passes before you turn on the triac, the 
    lamp will provide half intensity illumination.  So the PIC detects a zero 
    crossing, waits from 0 to 127 units of time, and then fires the triac.  This 
    process must be repeated for every half cycle.

    The PIC gets an interrupt every time a charactor is available from its' internal
    UART.  The program must save the charactors in a buffer as they are received.  The 
    program must then act on any messages whose address matches the address on its 
    configuration switches.  

    Special Notes, Bugs encountered (and fixed):
    -    
    -    
    - 
    - 
    - 

    To Do:

    Current Utilization:
        %ROM (Program  )........  37 %
        %RAM (Variables)........  66 %

******************************************************************************/
/*  I/O is as follows:                                                       */
/*     Pin                                                                   */
/*     .    Direction                                                        */
/*     .    .       Name / Description                                       */
/*---------------------------------------------------------------------------*/
/*     RA0  Out     Triac gate channel 1                                     */
/*     RA1  Out     Triac gate channel 2                                     */
/*     RA2  Out     Triac gate channel 3                                     */
/*     RA3  Out     Triac gate channel 4                                     */
/*     RA4  Out     Dry contact output                                       */
/*     RA5  Out     led                                                      */
/*     RB0  In      Control address switch bit 0                             */
/*     RB1  In      Control address switch bit 1                             */
/*     RB2  In      Control address switch bit 2                             */
/*     RB3  In      Control address switch bit 3                             */
/*     RB4  In      Control address switch bit 4                             */
/*     RB5  In      Control address switch bit 5                             */
/*     RB6  Unused                                                           */
/*     RB7  In      Zero crossing (interrupt on change capable pin)          */
/*     RC0  In      Midi channel switch bit 0                                */
/*     RC1  In      Midi channel switch bit 1                                */
/*     RC2  In      Midi channel switch bit 2                                */
/*     RC3  In      Midi channel switch bit 3                                */
/*     RC4  In      Dipswitch option 1                                       */
/*     RC5  In      Dipswitch option 2                                       */
/*     RC6  Unused                                                           */
/*     RC7  In      RX        Serial input, ASYNC                            */
/*****************************************************************************/
/*      TIMEBASE                                                             */
/*                                                                           */
/*      We'll use TIMER1 for our timebase.  A 20.00MHz Crystal will drive    */
/*      timer1.  Timer1 is a 16-bit counter, which can generate an interrupt */
/*      on overflow.  The Crystal oscillator goes first to a prescaler.      */
/*      We will set the prescaler to 4 for .2nS per count into the counter   */
/*      From there, it goes to the timer1 increment.  We preset the counter  */
/*      upon every overflow to 65208 to get an overflow every 328 counts     */
/*      So, the overflow should occur every 65 nano seconds (thats 1/127 of  */
/*      one half cycle of 60Hz AC.)                                          */
/*                                                                           */
/*      Four 8 bit software counters exist to keep up with each channels     */
/*      triac trigger delay.  A state machine exists in the timer interrupt  */
/*      routine that will trigger each channel as needed.                    */
/*                                                                           */
/*      60 Cycles per sec, 120 half cycles per second                        */
/*      1/120 = 8.3 mS per half cycle                                        */ 
/*      8.3mS / 127 possible brightness levels = 65.617nS per level          */
/*      65 / .2nS timer tick = 328 timer ticks per 127th of a half cycle     */
/*                                                                           */
/*****************************************************************************/

#include <16c63.h>

/* Tell compiler the crystal frequency.  This allows the 'delay_ms' to work as well as providing timing to the UART.*/
#use delay(clock=20000000)

/*  Assign direction to A,B and C ports.  Pins are by default an input, so we need to specify only the outputs. */
#use fixed_io(a_outputs= PIN_A0, PIN_A1, PIN_A2, PIN_A3, PIN_A4, PIN_A5)
#use standard_io(B)
#use standard_io(C)

#use rs232(baud=31250, xmit=PIN_C6, rcv=PIN_C7, PARITY=N) /* sets up the RS-232 to use the UART on pins C6 and C7 */

//#fuses   HS,NOPROTECT,WDT

/* this must be the first memory declaration, or the compiler will say 'out of ram' */
byte CONST PercentTable[130] = {
    0x7F,0x7E,0x7D,0x7C,0x7B,0x7A,0x79,0x78,0x77,0x76,0x75,0x74,0x73,0x72,0x71,0x70,
    0x6F,0x6E,0x6D,0x6C,0x6B,0x6A,0x69,0x68,0x67,0x66,0x65,0x64,0x63,0x62,0x61,0x60,
    0x5F,0x5E,0x5D,0x5C,0x5B,0x5A,0x59,0x58,0x57,0x56,0x55,0x54,0x53,0x52,0x51,0x50,
    0x4F,0x4E,0x4D,0x4C,0x4B,0x4A,0x49,0x48,0x47,0x46,0x45,0x44,0x43,0x42,0x41,0x40,
    0x3F,0x3E,0x3D,0x3C,0x3B,0x3A,0x39,0x38,0x37,0x36,0x35,0x34,0x33,0x32,0x31,0x30,
    0x2F,0x2E,0x2D,0x2C,0x2B,0x2A,0x29,0x28,0x27,0x26,0x25,0x24,0x23,0x22,0x21,0x20,
    0x1F,0x1E,0x1D,0x1C,0x1B,0x1A,0x19,0x18,0x17,0x16,0x15,0x14,0x13,0x12,0x11,0x10,
    0x0F,0x0E,0x0D,0x0C,0x0B,0x0A,0x09,0x08,0x07,0x06,0x05,0x04,0x03,0x02,0x01,0x01};


// int RequestedCh;              /* UART buffer to hold the channel number of the current MIDI request.              */
// int RequestedCntl;            /* UART buffer to hold the control number of the current MIDI request.              */
// int RequestedVal;             /* UART buffer to hold the control value of the current MIDI request.               */
int CntlOffset;                  /* which of the four triacs or relay is being addressed                             */
int UartRXbuf;                   /* UART buffer to hold the char in the RX buffer                                    */
int HaveRXchr;                   /* 1 when there's a MIDI char to be delt with                                       */
int MidiStatus;                  /* temp variable just for parsing MIDI Status byte                                  */
int Mychannel;                   /* The channel number that my dip switch is set to                                  */
int MyBaseCntlNum;               /* The base control number that my dip switch is set to                             */
int nStatepoll;                  /* state machine location                                                           */
int Triac1dly;                   /* num of AC ticcs to wait before triggering the triac after each zero crossing     */
int Triac2dly;                   /* num of AC ticcs to wait before triggering the triac after each zero crossing     */
int Triac3dly;                   /* num of AC ticcs to wait before triggering the triac after each zero crossing     */
int Triac4dly;                   /* num of AC ticcs to wait before triggering the triac after each zero crossing     */
int Triac1cnt;                   /* counts down from the delay value.  when it hits zero then trigger tirac          */
int Triac2cnt;                   /* counts down from the delay value.  when it hits zero then trigger tirac          */
int Triac3cnt;                   /* counts down from the delay value.  when it hits zero then trigger tirac          */
int Triac4cnt;                   /* counts down from the delay value.  when it hits zero then trigger tirac          */

int     portacopy;               /* holds a copy of what was last read from port A                                  */
int     portbcopy;               /* holds a copy of what was last read from port B                                  */
int     oldportbcopy;            /* holds a copy of what was read from port B last time thru                        */
int     oldportacopy;            /* holds a copy of what was read from port A last time thru                        */
#BYTE   PORTA=5                  /* Allows access to the whole port at one time                                     */
#BYTE   PORTB=6                  /* Allows access to the whole port at one time                                     */
#BYTE   PORTC=7                  /* Allows access to the whole port at one time                                     */
#BYTE   RCSTA=18                 /* (Hex address) status port for the UART                                          */

#define NotAddressedToMe 1        /* true when there's nothing of value in input buffer                              */ 
#define HaveChannelMatch 2        /* true when I get My channel number in buffer or its an "everybody listen"        */
#define HaveControlNum   3        /* true when I have the requested control number                                   */
#define JustGotChMatch   4        /* true when I just got a channel match with the char currently in the rxbuffer    */
#define RealTimeMsgChr   5        /* true when I just got a MIDI real time charactor                                 */
#define TRUE             1
#define FALSE            0 

void  InitializePIC             (void);   /* Set up basic hardware modes here                                  */
void  InterruptPortBChange      (void);   /* PORTB Chnge Interupt Svce Rutine                                  */
void  InterruptTimer1Overflow   (void);   /* Timer1 Overflow Interrupt                                         */
void  InterruptUartRX           (void);   /* Interrupt when a charactor appears in the RX buffer               */
 

main()
{
    InitializePIC ();                     /* Well, let's initialize stuff.  */
    
    Mychannel = (PORTC & 0x0F);          /* The channel number that my dip switch is set to       */
    MyBaseCntlNum = ((PORTB & 0x3F)*5);  /* The base control number that my dip switch is set to  */
                                         /* multiply by 5 because there's 5 controls per box      */

    /************************************************************************/   
    /*  Start Main Loop.                                                    */
    /*  The main loop is our basic embedded processor infinite loop.        */
 
    HaveRXchr = FALSE;
    nStatepoll = NotAddressedToMe;
    Triac1dly = Triac2dly = Triac3dly = Triac4dly = 0; 
    
    for (;;) 
    {
        /******** State Machine for handling MIDI word reception ************/
        if(HaveRXchr)
        {
            HaveRXchr = FALSE;                        /* if its a NoteOn or Control change command */                               
            MidiStatus = (UartRXbuf & 0xF0);
            switch(MidiStatus)
            {
                case 0x90:                                /* Note On                                      */
                case 0xB0:                                /* Control Change                               */
                    if( (UartRXbuf & 0x0F) == Mychannel ) /* if its commanding my channel                 */ 
                        nStatepoll = JustGotChMatch;      /* outside the switch so that a new address can */
                    else                                  /* override an old address                      */
                        nStatepoll = NotAddressedToMe;      
                break;
                case 0x80:                                /* Note Off                                     */
                case 0xA0:                                /* Polyphonic key pressure                      */
                case 0xC0:                                /* Program change                               */
                case 0xD0:                                /* Channel pressure                             */
                case 0xE0:                                /* pitch bend                                   */
                case 0xF0:                                /* System exclusive                             */

            }
 
            if( (UartRXbuf & 0xF0) == 0xF0)              /* if the MIDI char is a real time message   */
                ;                                        /* Ignore the charactor */
                
            if( (UartRXbuf & 0xF0) == 0x80)              /* if the MIDI char is a note off command    */
                nStatepoll = NotAddressedToMe;      
            if( (UartRXbuf & 0xF0) == 0xA0)              /* if the MIDI char is a key pressure cmd    */
                nStatepoll = NotAddressedToMe;      
  
            switch (nStatepoll)
            {
                case JustGotChMatch:
                    nStatepoll = HaveChannelMatch;         /* this state is used to throw the char away */
                break;   
                case NotAddressedToMe:                     /* this state is used to throw the char away */
                break;
                case HaveChannelMatch:
                    CntlOffset = (UartRXbuf & 0x3F) - (MyBaseCntlNum);
                    if( (CntlOffset >= 0 ) && (CntlOffset <= 4 ) )
                        nStatepoll = HaveControlNum;      /* now get the lamp brightess */
                    else
                        nStatepoll = NotAddressedToMe;      /* done everything that can be done... get set for next command */
                break;

                case HaveControlNum:
                    if(CntlOffset == 0) 
                        Triac1dly = PercentTable[(UartRXbuf & 0x7F)];     /* if dimmer 1 is being addressed */
                    if(CntlOffset == 1) 
                        Triac2dly = PercentTable[(UartRXbuf & 0x7F)];     /* if dimmer 2 is being addressed */
                    if(CntlOffset == 2) 
                        Triac3dly = PercentTable[(UartRXbuf & 0x7F)];     /* if dimmer 3 is being addressed */
                    if(CntlOffset == 3) 
                        Triac4dly = PercentTable[(UartRXbuf & 0x7F)];     /* if dimmer 4 is being addressed */
                    if(CntlOffset == 4) 
                        if(UartRXbuf & 0x40)                /* Turn relay on if value is greater than (dec)64 */
                            PORTA = portacopy = (portacopy | 0x10);   /* set pin A4 high to turn the relay on */
                        else                                         
                            PORTA = portacopy = (portacopy & 0xEF);   /* set pin A4 low to turn the relay off */
                     nStatepoll = NotAddressedToMe;      /* done everything that can be done... get set for next command */
                break;
            }
        }
    }
 }

void InitializePIC()  /* Call this once after reset.  Set up hardware configuration. */
{
    /* Setup TIMER1 so that it is a counter (not a timer), It will increment
       upon every rising edge of (system clk /4). The prescaler is set to 1. 
       so that the counter will increment every .2 uS (20MHz crystal) */

    set_tris_A(0);        /* setup port A to be outputs for the BYTE command */
    setup_timer_1 ( T1_INTERNAL | T1_DIV_BY_1);
    enable_interrupts (INT_TIMER1);        /* Enable timer 1 interrupt.      */
    enable_interrupts (INT_RDA);           /* Enable UART RX interrupt.      */
    enable_interrupts (RB_CHANGE);         /* Enable RB change interrupt.    */
    enable_interrupts (GLOBAL);            /* Allow interrupts in general    */
}

/*************************************************************************/
/*                                                                       */
/*     Interrupt Service Routines                                        */
/*                                                                       */
/*        Interrupts utilized are:                                       */
/*           1.  Timer1 Overflow                                         */
/*           2.  Port B Change                                           */
/*           3.  Uart receives a char                                    */
/*                                                                       */
/*************************************************************************/

#INT_TIMER1
void InterruptTimer1Overflow (void)
{
    set_timer1(65150);                    /* preset TIMER1 to rollover for 50Hz */
    /*set_timer1(65250);*/                /* preset TIMER1 to rollover in 285 counts for 60 Hz operation*/
    portacopy = (portacopy & 0xF0);       /* turn all triac triggers off      */
    if( --triac1cnt == 0 )                /* if its time to trigger the triac */
        portacopy = (portacopy | 0x01);   /* output_high(PIN_A0) triac 1 trig */
    if( --triac2cnt == 0 )                /* if its time to trigger the triac */
        portacopy = (portacopy | 0x02);   /* output_high(PIN_A1) triac 2 trig */
    if( --triac3cnt == 0 )                /* if its time to trigger the triac */
        portacopy = (portacopy | 0x04);   /* output_high(PIN_A2) triac 3 trig */
    if( --triac4cnt == 0 )                /* if its time to trigger the triac */
        portacopy = (portacopy | 0x08);   /* output_high(PIN_A3) triac 4 trig */
    /* If a triac gets triggered, it will pass current until AC cycle zero crosing */
    PORTA = portacopy;                    /* Set the PortA outputs */
}


/***  Interrupt Service Routine is called if any change on the B port ***/
/***  This resets the delays for each triac upon each zero crossing   ***/
#INT_RB
void InterruptPortBChange (void)
{
    oldportbcopy = portbcopy;    /* keep a copy of what was read the last time thru */
    portbcopy = PORTB;           /* read port B one time to clear interrupt */
    if( (portbcopy & 0x80) != (oldportbcopy & 0x80) )     /* if had a zero crossing */
    {
        Triac1cnt = triac1dly;       /* set a temp copy to count with   */
        Triac2cnt = triac2dly;       /* set a temp copy to count with   */
        Triac3cnt = triac3dly;       /* set a temp copy to count with   */
        Triac4cnt = triac4dly;       /* set a temp copy to count with   */
        if(portbcopy&0x80)           /* Turn led on if ac if going high */
            PORTA = portacopy = (portacopy | 0x20);   /* set pin A5 high to turn the led on */ 
        else                                         
            PORTA = portacopy = (portacopy & 0xDF);   /* set pin A5 low to turn the led off */

    }
}
#INT_RDA
void  InterruptUartRX (void)   /* Interrupt when a charactor appears in the RX buffer */
{
    UartRXbuf = getc();
    HaveRXchr = TRUE;
    if(bit_test(RCSTA,1))    /* if overrun bit is set */
    {
        bit_clear(RCSTA,4);  /* clear the overrun flag */
        bit_set(RCSTA,4);    /* set async rx mode enable */
    }
}


/*** END OF PROGRAM ***/






